//-----------------------------------------------------------------------------
//
// Copyright 2023 Intel Corporation All Rights Reserved.
//
// The source code contained or described herein and all documents related
// to the source code ("Material") are owned by Intel Corporation or its
// suppliers or licensors. Title to the Material remains with Intel
// Corporation or its suppliers and licensors. The Material contains trade
// secrets and proprietary and confidential information of Intel or its
// suppliers and licensors. The Material is protected by worldwide copyright
// and trade secret laws and treaty provisions. No part of the Material may
// be used, copied, reproduced, modified, published, uploaded, posted,
// transmitted, distributed, or disclosed in any way without Intel's prior
// express written permission.
//
// No license under any patent, copyright, trade secret or other intellectual
// property right is granted to or conferred upon you by disclosure or
// delivery of the Materials, either expressly, by implication, inducement,
// estoppel or otherwise. Any license under such intellectual property rights
// must be express and approved by Intel in writing.
//
//-----------------------------------------------------------------------------
// for initiator's formal verification:
//   - INITIATOR_TO_TARGET_AS_ASSERTS = 1
//   - TARGET_TO_INITIATOR_AS_ASSERTS = 0
// for target's formal verification:
//   - INITIATOR_TO_TARGET_AS_ASSERTS = 0
//   - TARGET_TO_INITIATOR_AS_ASSERTS = 1
// for simulation or formal verification at upper levels:
//   - INITIATOR_TO_TARGET_AS_ASSERTS = 1
//   - TARGET_TO_INITIATOR_AS_ASSERTS = 1
//-----------------------------------------------------------------------------

`ifdef ASSERTIONS_EN

module req_ack_interface_properties #(
  // configurable parameters:
  parameter INITIATOR_TO_TARGET_AS_ASSERTS  = 'x,                              // 1 -> properties for signals from initiator to target working as assertions
                                                                               // 0 -> properties for signals from initiator to target working as assumptions
  parameter TARGET_TO_INITIATOR_AS_ASSERTS  = 'x,                              // 1 -> properties for signals from target to initiator working as assertions
                                                                               // 0 -> properties for signals from target to initiator working as assumptions
  parameter REQ_DATA_WIDTH                  = 'x,                              // width of data sent with requests
  parameter ACK_DATA_WIDTH                  = 'x                               // width of data sent with acknowledgements
)(
  // clocks and resets:
  input logic                               clk,                               // clock
  input logic                               nreset,                            // asynchronous reset (active low)
  // request/acknowledgment interface:
  input logic                               req,                               // (initiator -> target) - request
  input logic          [REQ_DATA_WIDTH-1:0] req_data,                          // (initiator -> target) - request data
  input logic                               ack,                               // (target -> initiator) - acknowledgement
  input logic          [ACK_DATA_WIDTH-1:0] ack_data                           // (target -> initiator) - acknowledgement data
);

  //===========================================================================
  // internal signals
  //===========================================================================

  //---------------------------------------------------------------------------
  // subsidiary logic
  //---------------------------------------------------------------------------
  logic                                abv_time_after_cycle_after_rst_nxt_c;   // next state of the abv_time_after_cycle_after_rst_r register
  logic                                abv_time_after_cycle_after_rst_r;       // indicator of time after cycle after reset - register
  //-----
  logic                                abv_first_cycle_after_rst_c;            // indicator of first cycle after reset

  //===========================================================================
  // subsidiary logic
  //===========================================================================

  //---------------------------------------------------------------------------
  // indicator of time after cycle after reset - register
  //---------------------------------------------------------------------------
  // next state of the register:
  assign abv_time_after_cycle_after_rst_nxt_c = 1'b1;
  // register:
  always_ff @(posedge clk or negedge nreset)
    begin
    if (!nreset)
      abv_time_after_cycle_after_rst_r <= 1'b0;
    else
      abv_time_after_cycle_after_rst_r <= abv_time_after_cycle_after_rst_nxt_c;
    end

  //---------------------------------------------------------------------------
  // indicator of a first cycle after reset
  //---------------------------------------------------------------------------
  assign abv_first_cycle_after_rst_c = nreset &&
                                       !abv_time_after_cycle_after_rst_r;

  //===========================================================================
  // properties for signals from initiator to target
  //===========================================================================

  //---------------------------------------------------------------------------
  // request is always known
  //---------------------------------------------------------------------------
 `ifdef XPROP_EN
  property pr__x_req_always_known;
    @(posedge clk) disable iff (!nreset)
      (!$isunknown(req));
  endproperty
 `endif

  //---------------------------------------------------------------------------
  // request data is known, when request is high
  //---------------------------------------------------------------------------
 `ifdef XPROP_EN
  property pr__x_req_data_known_when_req_high;
    @(posedge clk) disable iff (!nreset)
      (req)
        |->
      (!$isunknown(req_data));
  endproperty
 `endif

  //---------------------------------------------------------------------------
  // request goes high only, when acknowledgement was low in a previous cycle
  //---------------------------------------------------------------------------
  property pr__req_goes_high_only_when_ack_low_in_prev_cycle;
    @(posedge clk) disable iff (!nreset)
      (((abv_first_cycle_after_rst_c &&
         req) ||
        (abv_time_after_cycle_after_rst_r &&
         $rose(req))))
        |->
      (abv_first_cycle_after_rst_c ||
       (abv_time_after_cycle_after_rst_r &&
        !$past(ack)));
  endproperty

  //---------------------------------------------------------------------------
  // request is kept high until acknowledgement goes high
  //---------------------------------------------------------------------------
  property pr__req_kept_high_until_ack_goes_high;
    @(posedge clk) disable iff (!nreset)
      (req &&
       !ack)
        |-> ##1
      (req);
  endproperty

  //---------------------------------------------------------------------------
  // after acknowledgement goes high, request ever goes low
  //---------------------------------------------------------------------------
  property pr__after_ack_goes_high_req_ever_goes_low;
    @(posedge clk) disable iff (!nreset)
      (abv_time_after_cycle_after_rst_r &&
       $rose(ack))
        |-> ##1
      (##[0:$] (!req));
  endproperty

  //---------------------------------------------------------------------------
  // request data is stable when request is high
  //---------------------------------------------------------------------------
  property pr__req_data_stable_when_req_high;
    @(posedge clk) disable iff (!nreset)
      (req)
        |-> ##1
      (!req ||
       $stable(req_data));
  endproperty

  //===========================================================================
  // assertions/assumes for signals from initiator to target
  //===========================================================================
  generate
    case (INITIATOR_TO_TARGET_AS_ASSERTS)
      //-----
      1: // assertions:
        begin : gb_initiator_to_target_asserts
 `ifdef XPROP_EN
        as__x_req_always_known                            : assert property(pr__x_req_always_known) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__x_req_data_known_when_req_high                : assert property(pr__x_req_data_known_when_req_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
 `endif
        as__req_goes_high_only_when_ack_low_in_prev_cycle : assert property(pr__req_goes_high_only_when_ack_low_in_prev_cycle) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__req_kept_high_until_ack_goes_high             : assert property(pr__req_kept_high_until_ack_goes_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__after_ack_goes_high_req_ever_goes_low         : assert property(pr__after_ack_goes_high_req_ever_goes_low) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__req_data_stable_when_req_high                 : assert property(pr__req_data_stable_when_req_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        end
      //-----
      0: // assumptions:
        begin : gb_initiator_to_target_assumes
 `ifdef XPROP_EN
        am__x_req_always_known                            : assume property(pr__x_req_always_known);
        am__x_req_data_known_when_req_high                : assume property(pr__x_req_data_known_when_req_high);
 `endif
        am__req_goes_high_only_when_ack_low_in_prev_cycle : assume property(pr__req_goes_high_only_when_ack_low_in_prev_cycle);
        am__req_kept_high_until_ack_goes_high             : assume property(pr__req_kept_high_until_ack_goes_high);
        am__after_ack_goes_high_req_ever_goes_low         : assume property(pr__after_ack_goes_high_req_ever_goes_low);
        am__req_data_stable_when_req_high                 : assume property(pr__req_data_stable_when_req_high);
        end
      //-----
    endcase
  endgenerate

  //===========================================================================
  // properties for signals from target to initiator
  //===========================================================================

  //---------------------------------------------------------------------------
  // acknowledgement is always known
  //---------------------------------------------------------------------------
 `ifdef XPROP_EN
  property pr__x_ack_always_known;
    @(posedge clk) disable iff (!nreset)
      (!$isunknown(ack));
  endproperty
 `endif

  //---------------------------------------------------------------------------
  // acknowledgement data is known, when acknowledgement is high
  //---------------------------------------------------------------------------
 `ifdef XPROP_EN
  property pr__x_ack_data_known_when_ack_high;
    @(posedge clk) disable iff (!nreset)
      (ack)
        |->
      (!$isunknown(ack_data));
  endproperty
 `endif

  //---------------------------------------------------------------------------
  // acknowledgement goes high only when request is high
  //---------------------------------------------------------------------------
  property pr__ack_goes_high_only_when_req_high;
    @(posedge clk) disable iff (!nreset)
      (((abv_first_cycle_after_rst_c &&
         ack) ||
        (abv_time_after_cycle_after_rst_r &&
         $rose(ack))))
        |->
      (req);
  endproperty

  //---------------------------------------------------------------------------
  // after request goes high, acknowledgement goes ever high
  //---------------------------------------------------------------------------
  property pr__after_req_goes_high_ack_ever_goes_high;
    @(posedge clk) disable iff (!nreset)
      (((abv_first_cycle_after_rst_c &&
         req) ||
        (abv_time_after_cycle_after_rst_r &&
         $rose(req))))
        |->
      (##[0:$] ack);
  endproperty

  //---------------------------------------------------------------------------
  // acknowledgement is kept high at least while request is high
  //---------------------------------------------------------------------------
  property pr__ack_kept_high_at_least_while_req_high;
    @(posedge clk) disable iff (!nreset)
      (ack &&
       req)
        |-> ##1
      (ack ||
       !req);
  endproperty

  //---------------------------------------------------------------------------
  // after request goes low, acknowledgement goes ever low
  //---------------------------------------------------------------------------
  property pr__after_req_goes_low_ack_goes_ever_low;
    @(posedge clk) disable iff (!nreset)
      (abv_time_after_cycle_after_rst_r &&
       $fell(req))
        |->
      (##[0:$] (!ack));
  endproperty

  //---------------------------------------------------------------------------
  // acknowledgement data is stable when acknowledgement is high
  //---------------------------------------------------------------------------
  property pr__ack_data_stable_when_ack_high;
    @(posedge clk) disable iff (!nreset)
      (ack)
        |-> ##1
      (!ack ||
       $stable(ack_data));
  endproperty

  //===========================================================================
  // assertions/assumes for signals from target to initiator
  //===========================================================================
  generate
    case (TARGET_TO_INITIATOR_AS_ASSERTS)
      //-----
      1: // assertions:
        begin : gb_target_to_initiator_asserts
 `ifdef XPROP_EN
        as__x_ack_always_known                     : assert property(pr__x_ack_always_known) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__x_ack_data_known_when_ack_high         : assert property(pr__x_ack_data_known_when_ack_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
 `endif
        as__ack_goes_high_only_when_req_high       : assert property(pr__ack_goes_high_only_when_req_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__after_req_goes_high_ack_ever_goes_high : assert property(pr__after_req_goes_high_ack_ever_goes_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__ack_kept_high_at_least_while_req_high  : assert property(pr__ack_kept_high_at_least_while_req_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__after_req_goes_low_ack_goes_ever_low   : assert property(pr__after_req_goes_low_ack_goes_ever_low) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__ack_data_stable_when_ack_high          : assert property(pr__ack_data_stable_when_ack_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        end
      //-----
      0: // assumptions:
        begin : gb_target_to_initiator_assumes
 `ifdef XPROP_EN
        am__x_ack_always_known                     : assume property(pr__x_ack_always_known);
        am__x_ack_data_known_when_ack_high         : assume property(pr__x_ack_data_known_when_ack_high);
 `endif
        am__ack_goes_high_only_when_req_high       : assume property(pr__ack_goes_high_only_when_req_high);
        am__after_req_goes_high_ack_ever_goes_high : assume property(pr__after_req_goes_high_ack_ever_goes_high);
        am__ack_kept_high_at_least_while_req_high  : assume property(pr__ack_kept_high_at_least_while_req_high);
        am__after_req_goes_low_ack_goes_ever_low   : assume property(pr__after_req_goes_low_ack_goes_ever_low);
        am__ack_data_stable_when_ack_high          : assume property(pr__ack_data_stable_when_ack_high);
        end
      //-----
    endcase
  endgenerate

endmodule

`endif
