//-----------------------------------------------------------------------------
//
// Copyright 2023 Intel Corporation All Rights Reserved.
//
// The source code contained or described herein and all documents related
// to the source code ("Material") are owned by Intel Corporation or its
// suppliers or licensors. Title to the Material remains with Intel
// Corporation or its suppliers and licensors. The Material contains trade
// secrets and proprietary and confidential information of Intel or its
// suppliers and licensors. The Material is protected by worldwide copyright
// and trade secret laws and treaty provisions. No part of the Material may
// be used, copied, reproduced, modified, published, uploaded, posted,
// transmitted, distributed, or disclosed in any way without Intel's prior
// express written permission.
//
// No license under any patent, copyright, trade secret or other intellectual
// property right is granted to or conferred upon you by disclosure or
// delivery of the Materials, either expressly, by implication, inducement,
// estoppel or otherwise. Any license under such intellectual property rights
// must be express and approved by Intel in writing.
//
//-----------------------------------------------------------------------------
// for initiator's formal verification:
//   - INITIATOR_TO_TARGET_AS_ASSERTS = 1
//   - TARGET_TO_INITIATOR_AS_ASSERTS = 0
// for target's formal verification:
//   - INITIATOR_TO_TARGET_AS_ASSERTS = 0
//   - TARGET_TO_INITIATOR_AS_ASSERTS = 1
// for simulation or formal verification at upper levels:
//   - INITIATOR_TO_TARGET_AS_ASSERTS = 1
//   - TARGET_TO_INITIATOR_AS_ASSERTS = 1
//-----------------------------------------------------------------------------

`ifdef ASSERTIONS_EN

module valid_ready_interface_properties #(
  // configurable parameters:
  parameter INITIATOR_TO_TARGET_AS_ASSERTS  = 'x,                              // 1 -> properties for signals from initiator to target working as assertions
                                                                               // 0 -> properties for signals from initiator to target working as assumptions
  parameter TARGET_TO_INITIATOR_AS_ASSERTS  = 'x,                              // 1 -> properties for signals from target to initiator working as assertions
                                                                               // 0 -> properties for signals from target to initiator working as assumptions
  parameter DATA_WIDTH                      = 'x                               // width of data
)(
  // clocks and resets:
  input logic                               clk,                               // clock
  input logic                               nreset,                            // asynchronous reset (active low)
  // valid/ready interface:
  input logic                               valid,                             // (initiator -> target) - valid flag
  input logic                               ready,                             // (target -> initiator) - ready flag
  input logic              [DATA_WIDTH-1:0] data                               // (initiator -> target) - data
);

  //===========================================================================
  // internal signals
  //===========================================================================

  //---------------------------------------------------------------------------
  // subsidiary logic
  //---------------------------------------------------------------------------
  logic                                abv_time_after_cycle_after_rst_nxt_c;   // next state of the abv_time_after_cycle_after_rst_r register
  logic                                abv_time_after_cycle_after_rst_r;       // indicator of time after cycle after reset - register
  //-----
  logic                                abv_first_cycle_after_rst_c;            // indicator of first cycle after reset

  //===========================================================================
  // subsidiary logic
  //===========================================================================

  //---------------------------------------------------------------------------
  // indicator of time after cycle after reset - register
  //---------------------------------------------------------------------------
  // next state of the register:
  assign abv_time_after_cycle_after_rst_nxt_c = 1'b1;
  // register:
  always_ff @(posedge clk or negedge nreset)
    begin
    if (!nreset)
      abv_time_after_cycle_after_rst_r <= 1'b0;
    else
      abv_time_after_cycle_after_rst_r <= abv_time_after_cycle_after_rst_nxt_c;
    end

  //---------------------------------------------------------------------------
  // indicator of a first cycle after reset
  //---------------------------------------------------------------------------
  assign abv_first_cycle_after_rst_c = nreset &&
                                       !abv_time_after_cycle_after_rst_r;

  //===========================================================================
  // properties for signals from initiator to target
  //===========================================================================

  //---------------------------------------------------------------------------
  // valid is always known
  //---------------------------------------------------------------------------
 `ifdef XPROP_EN
  property pr__x_valid_always_known;
    @(posedge clk) disable iff (!nreset)
      (!$isunknown(valid));
  endproperty
 `endif

  //---------------------------------------------------------------------------
  // data is known, when valid is high
  //---------------------------------------------------------------------------
 `ifdef XPROP_EN
  property pr__x_data_known_when_valid_high;
    @(posedge clk) disable iff (!nreset)
      (valid)
        |->
      (!$isunknown(data));
  endproperty
 `endif

  //---------------------------------------------------------------------------
  // valid is kept high, when ready is low
  //---------------------------------------------------------------------------
  property pr__valid_kept_high_when_ready_low;
    @(posedge clk) disable iff (!nreset)
      (valid &&
       !ready)
        |-> ##1
      (valid);
  endproperty

  //---------------------------------------------------------------------------
  // data is stable until ready is high
  //---------------------------------------------------------------------------
  property pr__data_stable_when_ready_low;
    @(posedge clk) disable iff (!nreset)
      (valid &&
       !ready)
        |-> ##1
      ($stable(data));
  endproperty

  //===========================================================================
  // assertions/assumes for signals from initiator to target
  //===========================================================================
  generate
    case (INITIATOR_TO_TARGET_AS_ASSERTS)
      //-----
      1: // assertions:
        begin : gb_initiator_to_target_asserts
 `ifdef XPROP_EN
        as__x_valid_always_known           : assert property(pr__x_valid_always_known) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__x_data_known_when_valid_high   : assert property(pr__x_data_known_when_valid_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
 `endif
        as__valid_kept_high_when_ready_low : assert property(pr__valid_kept_high_when_ready_low) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        as__data_stable_when_ready_low     : assert property(pr__data_stable_when_ready_low) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        end
      //-----
      0: // assumptions:
        begin : gb_initiator_to_target_assumes
 `ifdef XPROP_EN
        am__x_valid_always_known           : assume property(pr__x_valid_always_known);
        am__x_data_known_when_valid_high   : assume property(pr__x_data_known_when_valid_high);
 `endif
        am__valid_kept_high_when_ready_low : assume property(pr__valid_kept_high_when_ready_low);
        am__data_stable_when_ready_low     : assume property(pr__data_stable_when_ready_low);
        end
      //-----
    endcase
  endgenerate

  //===========================================================================
  // properties for signals from target to initiator
  //===========================================================================

  //---------------------------------------------------------------------------
  // ready is always known
  //---------------------------------------------------------------------------
 `ifdef XPROP_EN
  property pr__x_ready_always_known;
    @(posedge clk) disable iff (!nreset)
      (!$isunknown(ready));
  endproperty
 `endif

  //---------------------------------------------------------------------------
  // ready is ever high, when valid is high
  //---------------------------------------------------------------------------
  property pr__ready_ever_high_when_valid_high;
    @(posedge clk) disable iff (!nreset)
      ((abv_first_cycle_after_rst_c &&
        valid &&
        !ready) ||
       (abv_time_after_cycle_after_rst_r &&
        $rose(valid &&
              !ready)))
        |-> ##1
      (##[0:$] ready);
  endproperty

  //===========================================================================
  // assertions/assumes for signals from target to initiator
  //===========================================================================
  generate
    case (TARGET_TO_INITIATOR_AS_ASSERTS)
      //-----
      1: // assertions:
        begin : gb_target_to_initiator_asserts
 `ifdef XPROP_EN
        as__x_ready_always_known            : assert property(pr__x_ready_always_known) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
 `endif
        as__ready_ever_high_when_valid_high : assert property(pr__ready_ever_high_when_valid_high) else $error("%t: ERROR: ASSERTION FAILURE: %m", $time);
        end
      //-----
      0: // assumptions:
        begin : gb_target_to_initiator_assumes
 `ifdef XPROP_EN
        am__x_ready_always_known            : assume property(pr__x_ready_always_known);
 `endif
        am__ready_ever_high_when_valid_high : assume property(pr__ready_ever_high_when_valid_high);
        end
      //-----
    endcase
  endgenerate

endmodule

`endif
